/* Copyright (C) 2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Volker Friese [committer] */

#include "EventbuildChain.h"

#include "CbmDigiTimeslice.h"

#include <sstream>
#include <string>

#include "DigiData.h"
#include "evbuild/Config.h"

using namespace cbm::algo;
using namespace cbm::algo::evbuild;


// -----   Constructor   ------------------------------------------------------
EventbuildChain::EventbuildChain(const Config& config)
  : fTriggerDet(config.fTrigger.Detector())
  , fTrigger(config.fTrigger)
  , fBuilder(config.fBuilder)
  , fSelector(config.fSelector)
  , fQa(DigiEventQaConfig(config.fBuilder, 10., 100))
{
  Status();
}
// ----------------------------------------------------------------------------


// -----   Run event building on a timeslice   --------------------------------
EventbuildChain::ResultType EventbuildChain::Run(const DigiData& timeslice)
{

  // --- Extract the digi time stamps of the trigger detector
  std::vector<double> digiTimes = GetDigiTimes(timeslice, fTriggerDet);

  // --- Call the trigger algorithm
  auto [triggers, triggerMon] = fTrigger(digiTimes);

  // --- Perform event building
  auto [events, evbuildMon] = fBuilder(timeslice, triggers, fSelector);

  // --- Run event QA
  DigiEventQaData qaData = fQa(events);

  // --- Some log
  L_(info) << "Triggers: " << triggers.size() << ", events " << events.size();

  ResultType result;
  result.first          = std::move(events);
  result.second.trigger = triggerMon;
  result.second.evbuild = evbuildMon;

  return result;
}
// ----------------------------------------------------------------------------


// -----   Status info   ------------------------------------------------------
void EventbuildChain::Status() const
{
  L_(info) << "===   Eventbuiler configuration   ====================";
  L_(info) << "--- Using trigger detector " << ::ToString(fTriggerDet);
  L_(info) << fTrigger.ToString();
  L_(info) << fBuilder.ToString();
  L_(info) << fSelector.ToString();
  L_(info) << fQa.ToString();
  L_(info) << "======================================================";
}
// ----------------------------------------------------------------------------


// -----   Get digi times from CbmDigiTimeslice   -----------------------------
std::vector<double> EventbuildChain::GetDigiTimes(const DigiData& timeslice, ECbmModuleId system)
{
  std::vector<double> result;
  switch (system) {
    case ECbmModuleId::kSts: {
      result.resize(timeslice.fSts.size());
      auto it1 = timeslice.fSts.begin();
      auto it2 = timeslice.fSts.end();
      std::transform(it1, it2, result.begin(), [](const CbmStsDigi& digi) { return digi.GetTime(); });
      break;
    }
    case ECbmModuleId::kRich: {
      result.resize(timeslice.fRich.size());
      auto it1 = timeslice.fRich.begin();
      auto it2 = timeslice.fRich.end();
      std::transform(it1, it2, result.begin(), [](const CbmRichDigi& digi) { return digi.GetTime(); });
      break;
    }
    case ECbmModuleId::kMuch: {
      result.resize(timeslice.fMuch.size());
      auto it1 = timeslice.fMuch.begin();
      auto it2 = timeslice.fMuch.end();
      std::transform(it1, it2, result.begin(), [](const CbmMuchDigi& digi) { return digi.GetTime(); });
      break;
    }
    case ECbmModuleId::kTrd: {
      result.resize(timeslice.fTrd.size());
      auto it1 = timeslice.fTrd.begin();
      auto it2 = timeslice.fTrd.end();
      std::transform(it1, it2, result.begin(), [](const CbmTrdDigi& digi) { return digi.GetTime(); });
      break;
    }
    case ECbmModuleId::kTof: {
      result.resize(timeslice.fTof.size());
      auto it1 = timeslice.fTof.begin();
      auto it2 = timeslice.fTof.end();
      std::transform(it1, it2, result.begin(), [](const CbmTofDigi& digi) { return digi.GetTime(); });
      break;
    }
    case ECbmModuleId::kPsd: {
      result.resize(timeslice.fPsd.size());
      auto it1 = timeslice.fPsd.begin();
      auto it2 = timeslice.fPsd.end();
      std::transform(it1, it2, result.begin(), [](const CbmPsdDigi& digi) { return digi.GetTime(); });
      break;
    }
    case ECbmModuleId::kFsd: {
      result.resize(timeslice.fFsd.size());
      auto it1 = timeslice.fFsd.begin();
      auto it2 = timeslice.fFsd.end();
      std::transform(it1, it2, result.begin(), [](const CbmFsdDigi& digi) { return digi.GetTime(); });
      break;
    }
    case ECbmModuleId::kT0: {
      result.resize(timeslice.fBmon.size());
      auto it1 = timeslice.fBmon.begin();
      auto it2 = timeslice.fBmon.end();
      std::transform(it1, it2, result.begin(), [](const CbmTofDigi& digi) { return digi.GetTime(); });
      break;
    }
    default: {
      L_(error) << "EventbuildChain::GetDigiTimes: Unknown system " << system;
      break;
    }
  }  //? system

  return result;
}
// ----------------------------------------------------------------------------
