/* Copyright (C) 2023 FIAS Frankfurt Institute for Advanced Studies, Frankfurt / Main
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Felix Weiglhofer [committer] */
#include "Reco.h"

#include "CbmDigiEvent.h"

#include <Monitor.hpp>
#include <System.hpp>

#include <xpu/host.h>

#include "compat/OpenMP.h"
#include "config/Yaml.h"
#include "evbuild/Config.h"
#include "log.hpp"
#include "util/TimingsFormat.h"
#include "util/TsUtils.h"

using namespace cbm::algo;

Reco::Reco() {}
Reco::~Reco() {}

void Reco::Validate(const Options& opts)
{
  if (!fs::exists(opts.ParamsDir())) {
    throw std::runtime_error("ParamsDir does not exist: " + opts.ParamsDir().string());
  }

  if (opts.HasOutput(RecoData::Digi)) { throw std::runtime_error("Storing raw digis via 'Digi' option not supported"); }

  if (opts.HasOutput(RecoData::Cluster)) {
    throw std::runtime_error("Storing clusters via 'Cluster' option not supported");
  }
}

void Reco::Init(const Options& opts)
{
  if (fInitialized) throw std::runtime_error("Chain already initialized");

  Validate(opts);

  fContext.opts = opts;
  SetContext(&fContext);
  fUnpack.SetContext(&fContext);
  fStsHitFinder.SetContext(&fContext);
  fTracking.SetContext(&fContext);

  xpu::device_prop props {xpu::device::active()};
  L_(info) << "Running CBM Reco on Device '" << props.name() << "' (Using " << openmp::GetMaxThreads()
           << " OpenMP threads)";

  if (!opts.MonitorUri().empty()) {
    fContext.monitor = std::make_unique<cbm::Monitor>(opts.MonitorUri());
    L_(info) << "Monitoring enabled, sending to " << opts.MonitorUri();
  }

  // Reco Params
  fs::path recoParamsPath = opts.ParamsDir() / "RecoParams.yaml";
  YAML::Node yaml         = YAML::LoadFile(recoParamsPath.string());
  fContext.recoParams     = config::Read<RecoParams>(yaml);

  // Unpackers
  fUnpack.Init();

  // --- Event building
  fs::path configFile = opts.ParamsDir() / "EventbuildConfig.yaml";
  evbuild::Config config(YAML::LoadFile(configFile.string()));
  fEventBuild = std::make_unique<evbuild::EventbuildChain>(config);

  // STS Hitfinder
  fs::path stsHitfinderParamsPath    = opts.ParamsDir() / "StsHitfinder.yaml";
  yaml                               = YAML::LoadFile(stsHitfinderParamsPath.string());
  sts::HitfinderPars hitFinderConfig = config::Read<sts::HitfinderPars>(yaml);
  hitFinderConfig.landauTable        = sts::LandauTable::FromFile(opts.ParamsDir() / "LandauWidthTable.txt");
  fStsHitFinder.SetParameters(hitFinderConfig);

  // Tracking
  fTracking.Init();

  fInitialized = true;

  L_(debug) << "CBM Reco finished initialization";

  xpu::push_timer("Reco");
}


RecoResults Reco::Run(const fles::Timeslice& ts)
{
  if (!fInitialized) { throw std::runtime_error("Chain not initialized"); }

  for (uint64_t comp = 0; comp < ts.num_components(); comp++) {
    xpu::t_add_bytes(ts.size_component(comp));
  }

  xpu::push_timer(fmt::format("TS {}", ts.index()));

  L_(info) << ">>> Processing TS " << ts.index();
  xpu::set<cbm::algo::Params>(Params());

  Unpack::resultType unpackResult;
  UnpackMonitorData unpackMonitor;

  if (Opts().HasStep(Step::Unpack)) {
    switch (Params().sts.unpackMode) {
      case RecoParams::UnpackMode::XPU:
        // digis = fUnpackXpu.Exec(ts);
        throw std::runtime_error("XPU unpacker currently not implemented");
        break;
      default:
      case RecoParams::UnpackMode::CPU:
        unpackResult  = fUnpack.Run(ts);
        unpackMonitor = unpackResult.second;
        break;
    }
  }

  // --- Digi event building
  std::vector<DigiEvent> events;
  evbuild::EventbuildChainMonitorData evbuildMonitor;
  if (Opts().HasStep(Step::DigiTrigger)) {
    auto [ev, mon] = fEventBuild->Run(unpackResult.first);
    events         = std::move(ev);
    evbuildMonitor = mon;
  }

  sts::HitfinderMonitor stsHitfinderMonitor;
  PartitionedPODVector<sts::Hit> stsHits;
  if (Opts().HasStep(Step::LocalReco) && Opts().HasDetector(fles::Subsystem::STS)) {
    auto stsResults     = fStsHitFinder(unpackResult.first.fSts);
    stsHits             = std::move(stsResults.hits);
    stsHitfinderMonitor = stsResults.monitor;
  }

  // --- Tracking
  std::vector<ca::Track> tracks = fTracking.Run();

  xpu::timings ts_times = xpu::pop_timer();

  QueueUnpackerMetrics(ts, unpackMonitor, unpackResult.first);
  QueueStsRecoMetrics(stsHitfinderMonitor);
  QueueEvbuildMetrics(evbuildMonitor);

  PrintTimings(ts_times);

  RecoResults results;
  if (Opts().HasOutput(RecoData::DigiEvent)) results.events = std::move(events);
  if (Opts().HasOutput(RecoData::Hit)) results.stsHits = std::move(stsHits);

  return results;
}

void Reco::Finalize()
{
  fStsHitFinder.Finalize();
  fTracking.Finalize();

  // Pop timer that was started in Init()
  xpu::timings t = xpu::pop_timer();
  if (Opts().CollectKernelTimes()) {
    L_(info) << MakeReportSubtimers("Run Summary", fTimesliceTimesAcc) << "\n" << MakeReportSummary("Total", t);
  }
  else {
    L_(info) << "Total Processing time (Wall): " << t.wall() << " ms";
  }
}

void Reco::PrintTimings(xpu::timings& timings)
{
  if (Opts().CollectKernelTimes()) {
    L_(info) << MakeReportSubtimers("TS timings", timings) << "\n" << MakeReportSummary("Total", timings);
    fTimesliceTimesAcc.merge(timings);
  }
  else {
    L_(info) << "TS Processing time (Wall): " << timings.wall() << " ms";
  }
}

void Reco::QueueUnpackerMetrics(const fles::Timeslice& ts, const UnpackMonitorData& monitor, const DigiData& digis)
{
  if (!HasMonitor()) return;

  auto sizeBytes       = [](const auto& d) { return d.size() * sizeof(d[0]); };
  auto expansionFactor = [&](size_t x, const auto& d) {
    return d.size() > 0 ? static_cast<double>(sizeBytes(d)) / x : 0.0;
  };

  size_t tsDelta = ts.index() - prevTsId;
  prevTsId       = ts.index();

  auto& stsDigis   = digis.fSts;
  auto& muchDigis  = digis.fMuch;
  auto& tofDigis   = digis.fTof;
  auto& bmonDigis  = digis.fBmon;
  auto& trdDigis   = digis.fTrd;
  auto& trd2dDigis = digis.fTrd2d;
  auto& richDigis  = digis.fRich;

  size_t nDigisTotal = sizeBytes(stsDigis) + sizeBytes(muchDigis) + sizeBytes(tofDigis) + sizeBytes(bmonDigis)
                       + sizeBytes(trdDigis) + sizeBytes(trd2dDigis) + sizeBytes(richDigis);

  double totalExpansionFactor = static_cast<double>(nDigisTotal) / monitor.fNumBytes;

  GetMonitor().QueueMetric("cbmreco", {{"hostname", fles::system::current_hostname()}, {"child", Opts().ChildId()}},
                           {
                             {"tsIdDelta", tsDelta},
                             {"unpackTimeTotal", monitor.fTimeUnpack},
                             {"unpackBytesInSts", monitor.fNumBytesInSts},
                             {"unpackBytesInMuch", monitor.fNumBytesInMuch},
                             {"unpackBytesInTof", monitor.fNumBytesInTof},
                             {"unpackBytesInBmon", monitor.fNumBytesInBmon},
                             {"unpackBytesInTrd", monitor.fNumBytesInTrd},
                             {"unpackBytesInTrd2d", monitor.fNumBytesInTrd2d},
                             {"unpackBytesInRich", monitor.fNumBytesInRich},
                             {"unpackBytesOutSts", sizeBytes(stsDigis)},
                             {"unpackBytesOutMuch", sizeBytes(muchDigis)},
                             {"unpackBytesOutTof", sizeBytes(tofDigis)},
                             {"unpackBytesOutBmon", sizeBytes(bmonDigis)},
                             {"unpackBytesOutTrd", sizeBytes(trdDigis)},
                             {"unpackBytesOutTrd2d", sizeBytes(trd2dDigis)},
                             {"unpackBytesOutRich", sizeBytes(richDigis)},
                             {"unpackExpansionFactorSts", expansionFactor(monitor.fNumBytesInSts, stsDigis)},
                             {"unpackExpansionFactorMuch", expansionFactor(monitor.fNumBytesInMuch, muchDigis)},
                             {"unpackExpansionFactorTof", expansionFactor(monitor.fNumBytesInTof, tofDigis)},
                             {"unpackExpansionFactorBmon", expansionFactor(monitor.fNumBytesInBmon, bmonDigis)},
                             {"unpackExpansionFactorTrd", expansionFactor(monitor.fNumBytesInTrd, trdDigis)},
                             {"unpackExpansionFactorTrd2d", expansionFactor(monitor.fNumBytesInTrd2d, trd2dDigis)},
                             {"unpackExpansionFactorRich", expansionFactor(monitor.fNumBytesInRich, richDigis)},
                             {"unpackExpansionFactorTotal", totalExpansionFactor},
                             {"unpackNumCompUsed", monitor.fNumCompUsed},
                             {"unpackNumErrInvalidEqId", monitor.fNumErrInvalidEqId},
                             {"unpackNumErrInvalidSysVer", monitor.fNumErrInvalidSysVer},
                           });
}

void Reco::QueueStsRecoMetrics(const sts::HitfinderMonitor& monitor)
{
  if (!HasMonitor()) return;

  GetMonitor().QueueMetric("cbmreco", {{"hostname", fles::system::current_hostname()}, {"child", Opts().ChildId()}},
                           {
                             {"stsRecoTimeTotal", monitor.fTimeTotal},
                             {"stsRecoNumClusters", (unsigned long) monitor.fNumClusterTotal},
                             {"stsRecoNumHits", (unsigned long) monitor.fNumHitsTotal},
                             {"stsRecoNumClusterBucketOverflow", monitor.fNumClusterBucketOverflow},
                             {"stsRecoNumHitBucketOverflow", monitor.fNumHitBucketOverflow},
                           });
}

void Reco::QueueEvbuildMetrics(const evbuild::EventbuildChainMonitorData& mon)
{
  if (!HasMonitor()) return;

  const MetricTagSet tags = {{"hostname", fles::system::current_hostname()}, {"child", Opts().ChildId()}};

  size_t nDigisTotal         = 0;
  size_t nDigisInEventsTotal = 0;

  auto queueDetMetrics = [&](std::string_view det, auto& detMon) {
    size_t nDigis         = detMon.nDigis;
    size_t nDigisInEvents = detMon.nDigisInEvents;
    double selectionRatio = nDigis > 0 ? double(nDigisInEvents) / nDigis : 0;

    nDigisTotal += nDigis;
    nDigisInEventsTotal += nDigisInEvents;

    GetMonitor().QueueMetric("cbmreco", tags,
                             {{fmt::format("{}NumDigisTotal", det), nDigis},
                              {fmt::format("{}NumDigisInEvents", det), nDigisInEvents},
                              {fmt::format("{}EvSelectionRatio", det), selectionRatio}});
  };

  queueDetMetrics("sts", mon.evbuild.sts);
  queueDetMetrics("much", mon.evbuild.much);
  queueDetMetrics("tof", mon.evbuild.tof);
  queueDetMetrics("bmon", mon.evbuild.bmon);
  queueDetMetrics("trd", mon.evbuild.trd);
  queueDetMetrics("trd2d", mon.evbuild.trd2d);
  queueDetMetrics("rich", mon.evbuild.rich);

  double totalSelectionRatio = nDigisTotal > 0 ? double(nDigisInEventsTotal) / nDigisTotal : 0;
  GetMonitor().QueueMetric("cbmreco", tags,
                           {{"digiTriggerTimeTotal", mon.trigger.timeMs},
                            {"eventbuildTimeTotal", mon.evbuild.timeMs},
                            {"numTrigger", mon.trigger.nTriggers},
                            {"numEvents", mon.evbuild.nEvents},
                            {"totalEvSelectionRatio", totalSelectionRatio}});
}
