/* Copyright (C) 2023 FIAS Frankfurt Institute for Advanced Studies, Frankfurt / Main
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Felix Weiglhofer [committer] */

#include "PODVector.h"
#include "PartitionedVector.h"
#include "gtest/gtest.h"

using namespace cbm::algo;

template<typename T>
void EXPECT_CONTAINER_EQ(gsl::span<T> a, std::vector<T> b)
{
  EXPECT_EQ(a.size(), b.size());
  for (size_t i = 0; i < a.size(); ++i) {
    EXPECT_EQ(a[i], b[i]);
  }
}

TEST(PartitionedVector, CanCreateEmpty)
{
  PartitionedVector<i32> vec;
  EXPECT_EQ(vec.NElements(), 0);
  EXPECT_EQ(vec.NPartitions(), 0);
}

TEST(PartitionedVector, CanCreateWithPartitions)
{
  std::vector<i32> data      = {1, 2, 3, 4, 5, 6, 7, 8, 9};
  std::vector<size_t> sizes  = {2, 4, 3};
  std::vector<u32> addresses = {0x0, 0x100, 0x200};

  PartitionedVector vec(std::move(data), sizes, addresses);

  EXPECT_EQ(vec.NElements(), 9);
  EXPECT_EQ(vec.NPartitions(), 3);

  EXPECT_EQ(vec.Size(0), 2);
  EXPECT_EQ(vec[0].size(), 2);
  EXPECT_EQ(vec.Address(0), 0x0);
  EXPECT_CONTAINER_EQ(vec[0], {1, 2});

  auto part = vec.Partition(0);
  EXPECT_EQ(part.first.size(), vec.Size(0));
  EXPECT_EQ(part.second, vec.Address(0));

  EXPECT_EQ(vec.Size(1), 4);
  EXPECT_EQ(vec[1].size(), 4);
  EXPECT_EQ(vec.Address(1), 0x100);
  EXPECT_CONTAINER_EQ(vec[1], {3, 4, 5, 6});

  part = vec.Partition(1);
  EXPECT_EQ(part.first.size(), vec.Size(1));
  EXPECT_EQ(part.second, vec.Address(1));

  EXPECT_EQ(vec.Size(2), 3);
  EXPECT_EQ(vec[2].size(), 3);
  EXPECT_EQ(vec.Address(2), 0x200);
  EXPECT_CONTAINER_EQ(vec[2], {7, 8, 9});

  part = vec.Partition(2);
  EXPECT_EQ(part.first.size(), vec.Size(2));
  EXPECT_EQ(part.second, vec.Address(2));
}

TEST(PartitionedVector, ThrowsOnAddressSizesMismatch)
{
  std::vector<i32> data      = {1, 2, 3, 4, 5, 6, 7, 8, 9};
  std::vector<size_t> sizes  = {2, 4, 3};
  std::vector<u32> addresses = {0x0, 0x100};

  EXPECT_THROW(PartitionedVector vec(std::move(data), sizes, addresses), std::runtime_error);
}

TEST(PartitionedVector, ThrowsOnSizesMismatch)
{
  std::vector<i32> data      = {1, 2, 3, 4, 5, 6, 7, 8, 9};
  std::vector<size_t> sizes  = {2, 3, 3};
  std::vector<u32> addresses = {0x0, 0x100, 0x200};

  EXPECT_THROW(PartitionedVector vec(std::move(data), sizes, addresses), std::runtime_error);
}

TEST(PartitionedVector, ThrowsOnOutOfBounds)
{
  std::vector<i32> data      = {1, 2, 3, 4, 5, 6, 7, 8, 9};
  std::vector<size_t> sizes  = {2, 4, 3};
  std::vector<u32> addresses = {0x0, 0x100, 0x200};

  PartitionedVector vec(std::move(data), sizes, addresses);

  EXPECT_THROW(vec[3], std::out_of_range);
  EXPECT_THROW(vec.Partition(3), std::out_of_range);
  EXPECT_THROW(vec.Address(3), std::out_of_range);
  EXPECT_THROW(vec.Size(3), std::out_of_range);
}

TEST(PartitionedVector, IsCopyConstructable) {}
