/* Copyright (C) 2023 FIAS Frankfurt Institute for Advanced Studies, Frankfurt / Main
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Felix Weiglhofer [committer] */
#include "CbmDigiEvent.h"

#include <TimesliceAutoSource.hpp>

#include <log.hpp>
#include <sstream>

#include <xpu/host.h>

#include "BuildInfo.h"
#include "Options.h"
#include "Reco.h"
#include "RecoResultsInputArchive.h"
#include "RecoResultsOutputArchive.h"
#include "compat/OpenMP.h"

using namespace cbm::algo;

std::shared_ptr<StorableRecoResults> makeStorableRecoResults(const fles::Timeslice& ts, const RecoResults& results)
{
  auto storable = std::make_shared<StorableRecoResults>(ts.index(), ts.start_time());

  storable->DigiEvents().reserve(results.events.size());
  for (const auto& digiEvent : results.events) {
    storable->DigiEvents().emplace_back(digiEvent.ToStorable());
  }

  storable->StsHits() = results.stsHits;

  return storable;
}

bool dumpArchive(const Options& opts)
{
  // Limit the number of events per timeslice to dump to avoid spamming the log
  constexpr size_t DumpEventsPerTS = 10;
  constexpr size_t DumpHitsPerSensor = 2;

  if (!opts.DumpArchive()) return false;

  L_(info) << "Dumping archive: " << opts.InputLocator();

  RecoResultsInputArchive archive(opts.InputLocator());

  auto desc = archive.descriptor();
  L_(info) << "Archive descriptor: ";
  L_(info) << " - time_created: " << desc.time_created();
  L_(info) << " - hostname: " << desc.hostname();
  L_(info) << " - username: " << desc.username();

  for (auto recoResults = archive.get(); !archive.eos(); recoResults = archive.get()) {
    if (recoResults == nullptr) {
      L_(error) << "Failed to read RecoResults from archive";
      break;
    }

    size_t nEvents = recoResults->DigiEvents().size();
    L_(info) << "TS " << recoResults->TsIndex() << " start: " << recoResults->TsStartTime() << " events: " << nEvents
             << ", stsHits: " << recoResults->StsHits().NElements();

    for (size_t i = 0; i < std::min(DumpEventsPerTS, nEvents); i++) {
      const auto& digiEvent = recoResults->DigiEvents().at(i);
      L_(info) << " - Event " << i << " number: " << digiEvent.fNumber << "; time: " << digiEvent.fTime
               << "; nStsDigis: " << digiEvent.fData.Size(ECbmModuleId::kSts)
               << "; nTofDigis: " << digiEvent.fData.Size(ECbmModuleId::kTof);
    }

    if (nEvents > DumpEventsPerTS) L_(info) << "...";

    auto& stsHits = recoResults->StsHits();
    for (size_t m = 0; m < stsHits.NPartitions(); m++) {
      auto [hits, address] = stsHits.Partition(m);
      for (size_t i = 0; i < std::min(DumpHitsPerSensor, hits.size()); i++) {
        const auto& hit = hits[i];
        L_(info) << " - STS Hit " << i << " sensor: " << address << "; time: " << hit.fTime << ", X: " << hit.fX
                 << ", Y: " << hit.fY << ", Z: " << hit.fZ;
      }
    }


    if (nEvents > DumpEventsPerTS) L_(info) << "...";
  }

  return true;
}

int main(int argc, char** argv)
{
  Options opts(argc, argv);

  logging::add_console(opts.LogLevel());

  if (!opts.LogFile().empty()) logging::add_file(opts.LogFile().string(), opts.LogLevel());

  // XPU
  xpu::settings settings;
  settings.profile = opts.CollectKernelTimes();
  settings.device  = opts.Device();
  if (opts.LogLevel() == trace) {
    settings.verbose      = true;
    settings.logging_sink = [](std::string_view msg) { L_(trace) << msg; };
  }
  xpu::initialize(settings);
  xpu::preload<GPUReco>();

  auto ompThreads = opts.NumOMPThreads();
  if (ompThreads) {
    L_(debug) << *ompThreads << " OpenMP threads requested";
    openmp::SetNumThreads(*ompThreads);
  }

  L_(info) << "CBMRECO buildType=" << BuildInfo::BUILD_TYPE << " gpuDebug=" << BuildInfo::GPU_DEBUG
           << " parallelSTL=" << BuildInfo::WITH_PARALLEL_ALGORITHM << " OMP=" << BuildInfo::WITH_OMP
           << " commit=" << BuildInfo::GIT_HASH;
  std::stringstream ss;
  for (int i = 0; i < argc; i++) {
    ss << argv[i] << " ";
  }
  L_(info) << ss.str();

  if (dumpArchive(opts)) return 0;

  Reco reco;
  reco.Init(opts);

  fles::TimesliceAutoSource source(opts.InputLocator());

  std::optional<RecoResultsOutputArchive> archive;
  if (!opts.OutputFile().empty()) {
    L_(info) << "Writing results to file: " << opts.OutputFile();
    archive.emplace(opts.OutputFile().string());
  }

  int tsIdx  = 0;
  int num_ts = opts.NumTimeslices();
  if (num_ts > 0) num_ts += opts.SkipTimeslices();
  L_(debug) << "Starting to fetch timeslices from source...";
  while (auto ts = source.get()) {
    if (tsIdx < opts.SkipTimeslices()) {
      tsIdx++;
      continue;
    }
    auto result = reco.Run(*ts);

    if (archive) {
      auto storable = makeStorableRecoResults(*ts, result);
      archive->put(storable);
    }
    tsIdx++;

    if (num_ts > 0 && tsIdx >= num_ts) break;
  }

  if (archive) archive->end_stream();

  reco.Finalize();

  return 0;
}
